/***********************************************************************
 *
 *      Checkers v2.2 (c) David A.S.Crofts
 *
 *      23 Brisbane Street
 *      Berwick Victoria 3806
 *      Australia
 *
 *      email: dasc1961@netscape.net
 *
 ***********************************************************************/

#include <PalmOS.h>

#include "checkers.h"
#include "checkersrc.h"


/***********************************************************************
 *
 *      table.c - External Variables and Constants
 *
 ***********************************************************************/

extern UInt8 num_moves[4][32];
extern UInt8 moves[4][32][4];
extern UInt8 jumps[4][32][4];
extern PIECE_NOS computer_turn_nos;
extern PIECE_NOS oponent_turn_nos;

extern PIECE_NOS_PTR t;
extern Int16 jump_found;


/***********************************************************************
 *
 *      Internal Constants
 *
 ***********************************************************************/

#define BoardOffsetX    16
#define BoardOffsetY    17

#define BitmapWidth     16
#define BitmapHeight    16


/***********************************************************************
 *
 *      Private global variables
 *
 ***********************************************************************/

static GameType Game;
static CheckersPrefType CheckersPref;

static WinHandle OffscreenBoardWinH = 0;
static WinHandle OffscreenBitmapWinH = 0;

static Int16 BitmapTable[] = {
    GreenSquareEmptyBitmap,
    WhiteSquareEmptyBitmap,
    WhiteSquareRedManBitmap,
    WhiteSquareRedKingBitmap,
    WhiteSquareBlackManBitmap,
    WhiteSquareBlackKingBitmap,
    HiLightedSquareEmptyBitmap,
    HiLightedSquareRedManBitmap,
    HiLightedSquareRedKingBitmap,
    MoveEndSquareBitmap,
    MoveMidSquareBitmap,
    ComputerStartSquareBitmap,
    ComputerEndSquareManBitmap,
    ComputerEndSquareKingBitmap,
    ComputerJumpedSquareBitmap
};

static STATE new_move, move, old_move, start;

static STATE_PTR FreeStates = NULL;
static STATE_PTR HeadPrevStates = NULL;
static STATE_PTR HeadNextStates = NULL;

static SUCESSOR_PTR FreeSucessors = NULL;
static SUCESSOR_PTR r = NULL;


/***********************************************************************
 *
 *      Internal Functions
 *
 ***********************************************************************/

static UInt16 StartApplication (void);
static void StopApplication (void);
static Boolean MainFormDoCommand (UInt16 command);
static void MainFormInit (FormPtr frm);
static Boolean MainFormHandleEvent (EventPtr event);
static Boolean CurrentFormHandleEvent (EventPtr event);

static void DrawGameBoard (void);
static void UpdateStatusDisplay (void);
static Boolean HandlePenDown (Int16 penX, Int16 penY, Boolean marking,
                              Boolean * inBoundsP);
static Boolean MapPenPosition (Int16 penX, Int16 penY,
                               PieceCoordType * coordP);
static void PointerMove (Int16 i);
static void RemoveStateFlags (STATE_PTR p);
static void MakeComputerPlay (void);
static void PostMoveRoutine (void);
static void DrawPiece (WinHandle dstWinH, Int16 row, Int16 col);
static void DrawBitmap (WinHandle winHandle, Int16 resID, Int16 x, Int16 y);
static void GetPieceRectangle (Int16 row, Int16 col, RectanglePtr rP);
static void NewGameBoard (void);
static void SaveGameBoard (void);
static void LoadGameBoard (void);
static Int16 ConvertSquare (UInt16 bitmap);
static Int16 ConvertBitmap (UInt16 piece);
static void CaptureGraphics (void);
static void RestoreGraphics (void);

static SUCESSOR_PTR play (STATE_PTR p, Int16 depth);
static SUCESSOR_PTR gen_sucessors (STATE_PTR p);
static SUCESSOR_PTR gen_single_step_sucessors (STATE_PTR p);
static SUCESSOR_PTR consider_jumps (STATE_PTR p);
static SUCESSOR_PTR New_SUCESSOR (void);
static STATE_PTR New_STATE (void);
static Int16 value (STATE_PTR p);
static Boolean states_equal (STATE_PTR p, STATE_PTR q);
static void copy_state (STATE_PTR p, STATE_PTR q);
static void copy_state_play (STATE_PTR p, STATE_PTR q);
static void PushState (STATE_PTR * r, STATE_PTR q);
static void EmptyFreeLists (void);
static void free_sucessors (SUCESSOR_PTR q);
static void end_game_tweek (SUCESSOR_PTR r);
static void set_clump_value (STATE_PTR p);
static void TrimHistory (void);

Int16 f (STATE_PTR p, Int16 alpha, Int16 beta, Int16 depth);
void add_jumps (SUCESSOR_PTR q, STATE_PTR p, Int16 jumped);


/***********************************************************************
 *
 * FUNCTION:     NewGameBoard
 *
 ***********************************************************************/

static void
NewGameBoard (void)
{
    UInt16 i;
    STATE_PTR p;

    // Initialize the board

    Game.restored = false;
    Game.state = gameInProgress;
    Game.value = 0;
    Game.temp_depth = Game.depth;
    Game.draw = false;
    Game.first_move = true;
    Game.game_over = false;
    Game.player_jump_found = false;
    Game.compulsory_jump_found = false;
    Game.default_piece = NO_PIECE;

    for (i = 0; i < 64; ++i) {
        Game.square[i] = GreenSquareEmptyBitmap;
    }
    for (i = 0; i < 8; i = i + 2) {
        Game.square[i] = WhiteSquareRedManBitmap;
    }
    for (i = 8; i < 16; i = i + 2) {
        Game.square[i + 1] = WhiteSquareRedManBitmap;
    }
    for (i = 16; i < 24; i = i + 2) {
        Game.square[i] = WhiteSquareRedManBitmap;
    }
    for (i = 24; i < 32; i = i + 2) {
        Game.square[i + 1] = WhiteSquareEmptyBitmap;
    }
    for (i = 32; i < 40; i = i + 2) {
        Game.square[i] = WhiteSquareEmptyBitmap;
    }
    for (i = 40; i < 48; i = i + 2) {
        Game.square[i + 1] = WhiteSquareBlackManBitmap;
    }
    for (i = 48; i < 56; i = i + 2) {
        Game.square[i] = WhiteSquareBlackManBitmap;
    }
    for (i = 56; i < 64; i = i + 2) {
        Game.square[i + 1] = WhiteSquareBlackManBitmap;
    }
    for (i = 0; i < 64; ++i) {
        Game.new_square[i] = Game.square[i];
        Game.old_square[i] = Game.square[i];
    }

    CaptureGraphics ();

    while (HeadPrevStates != NULL) {
        p = HeadPrevStates;
        HeadPrevStates = HeadPrevStates->memPtr;
        MemHandleFree (p->memHandle);
    }

    PushState (&HeadPrevStates, &new_move);

}


/***********************************************************************
 *
 * FUNCTION:     LoadGameBoard
 *
 ***********************************************************************/

static void
LoadGameBoard (void)
{
    Int16 i;
    Boolean found;

    // Try loading a saved game board - if that fails, generate a new one


    for (i = 0; i < 12; ++i) {
        start.board[i] = PLAYER_PIECE;
    }
    for (i = 12; i < 20; ++i) {
        start.board[i] = NO_PIECE;
    }
    for (i = 20; i < 32; ++i) {
        start.board[i] = COMPUTER_PIECE;
    }

    found =
        PrefGetAppPreferencesV10 (checkersAppCreator, checkersAppVersionNum,
                                  &CheckersPref, sizeof (CheckersPref));

    if (found &&
        CheckersPref.signature == CheckersPrefSignature &&
        CheckersPref.game.state == gameInProgress) {
        Game = CheckersPref.game;
        Game.restored = true;

        for (i = 0; i < 64; ++i) {
            Game.new_square[i] = Game.square[i];
        }

        Game.default_piece = NO_PIECE;

        CaptureGraphics ();

        if (states_equal (&new_move, &start))
            PushState (&HeadPrevStates, &new_move);
        else {
            PushState (&HeadPrevStates, &new_move);
            PushState (&HeadPrevStates, &new_move);
        }

        if (Game.compulsory_jump_found)
            PointerMove (Game.compulsory_jump_start_square);

    }
    else {
        Game.depth = 6;
        Game.restored = false;
        NewGameBoard ();
    }

}


/***********************************************************************
 *
 * FUNCTION:     SaveGameBoard
 *
 ***********************************************************************/

static void
SaveGameBoard (void)
{
    CheckersPref.signature = CheckersPrefSignature;
    CheckersPref.game = Game;

    // Save our preferences to the Preferences database

    PrefSetAppPreferencesV10 (checkersAppCreator,
                              checkersAppVersionNum,
                              &CheckersPref, sizeof (CheckersPref));
}


/***********************************************************************
 *
 * FUNCTION:     HandlePenDown
 *
 ***********************************************************************/

static Boolean
HandlePenDown (Int16 penX, Int16 penY, Boolean marking, Boolean * inBoundsP)
{
    PieceCoordType coord;
    Int16 i, j, index;
    Boolean handled;
    STATE_PTR p;

    // Map pen position to game board row and column

    *inBoundsP = MapPenPosition (penX, penY, &coord);

    // If not in game board bounds or game is over, bail out

    if (!(*inBoundsP) || Game.state != gameInProgress)
        return (false);

    i = ((7 - coord.row) * 8) + coord.col;

    if (Game.new_square[i] == GreenSquareEmptyBitmap)
        return (false);

    handled = false;

    i = ((7 - coord.row) * 4) + 3 - (coord.col / 2);

    switch (new_move.board[i]) {

    case PLAYER_PIECE:
    case PLAYER_KING:
        if (Game.game_over)
            break;
        if (Game.default_piece != NO_PIECE && !Game.compulsory_jump_found) {
            for (j = 0; j < 32; ++j) {
                if (new_move.board[j] == HI_LIGHTED_NO_PIECE) {
                    new_move.board[j] = Game.default_piece;
                    break;
                }
            }
            RemoveStateFlags (&new_move);
            Game.default_piece = NO_PIECE;
        }
        Game.default_piece = new_move.board[i];
        if (!Game.compulsory_jump_found) {
            PointerMove (i);
            if (new_move.board[i] == PLAYER_PIECE ||
                new_move.board[i] == PLAYER_KING) {
                handled = true;
                break;
            }
        }
        else {
            break;
        }

        handled = true;
        break;

    case HI_LIGHTED_PLAYER_PIECE:
    case HI_LIGHTED_PLAYER_KING:
        Game.default_piece = new_move.board[i];
        new_move.board[i] = HI_LIGHTED_NO_PIECE;
        handled = true;
        break;

    case HI_LIGHTED_END_SQUARE:
        Game.default_piece = NO_PIECE;
        for (j = 0; j < (Int16) r->num; ++j) {
            if (r->ptr[j]->piece == i) {
                copy_state (r->ptr[j], &new_move);
                break;
            }
        }
        RestoreGraphics ();
        DrawGameBoard ();
        MakeComputerPlay ();
        return (true);

    case HI_LIGHTED_MID_SQUARE:
        for (j = 0; j < (Int16) r->num; ++j) {
            if (r->ptr[j]->piece - 128 == i) {
                copy_state (r->ptr[j], &new_move);
                break;
            }
        }
        PointerMove (i);
        handled = true;
        break;

    case HI_LIGHTED_NO_PIECE:
        if (!Game.compulsory_jump_found) {
            new_move.board[i] = Game.default_piece;
            RemoveStateFlags (&new_move);
            Game.default_piece = NO_PIECE;
        }
        handled = true;
        break;

    default:
        break;
    }

    if (handled) {
        RestoreGraphics ();
        DrawGameBoard ();
    }

    return (handled);
}


/***********************************************************************
 *
 * FUNCTION:     MapPenPosition
 *
 ***********************************************************************/

static Boolean
MapPenPosition (Int16 penX, Int16 penY, PieceCoordType * coordP)
{
    Int16 x;
    Int16 y;
    RectangleType rect;

    // Map display relative coordinates to window-relative

    x = (Int16) penX;
    y = (Int16) penY;

    WinDisplayToWindowPt (&x, &y);

    rect.topLeft.x = BoardOffsetX;
    rect.topLeft.y = BoardOffsetY;
    rect.extent.x = 8 * BitmapWidth;
    rect.extent.y = 8 * BitmapHeight;

    if (!RctPtInRectangle (x, y, &rect))
        return (false);

    // Convert to board-relative coordinates

    x -= BoardOffsetX;
    y -= BoardOffsetY;

    if (x < 0) {
        ErrDisplay ("board x is negative");
        x = 0;
    }

    if (y < 0) {
        ErrDisplay ("board y is negative");
        y = 0;
    }

    // Get the piece position

    coordP->col = x / BitmapWidth;
    coordP->row = y / BitmapHeight;

    if (coordP->col > 7) {
        ErrDisplay ("column out of bounds");
        coordP->col = 7;
    }
    if (coordP->row > 7) {
        ErrDisplay ("row out of bounds");
        coordP->row = 7;
    }

    return (true);
}


/***********************************************************************
 *
 * FUNCTION:     DrawPiece
 *
 ***********************************************************************/

static void
DrawPiece (WinHandle dstWinH, Int16 row, Int16 col)
{
    RectangleType rect;

    Int16 bitmapID = 0;
    Int16 bitmapIndex = 0;

    bitmapID = Game.new_square[(7 - row) * 8 + col];
    bitmapIndex = bitmapID - 1000;

    // Get bounds of the game piece

    GetPieceRectangle (row, col, &rect);

    // Draw the bitmap

    // Optimization

    if (OffscreenBitmapWinH) {

        RectangleType srcRect;

        srcRect.topLeft.x = BitmapWidth * bitmapIndex;
        srcRect.topLeft.y = 0;
        srcRect.extent.x = BitmapWidth;
        srcRect.extent.y = BitmapHeight;
        WinCopyRectangle (OffscreenBitmapWinH, dstWinH, &srcRect,
                          rect.topLeft.x, rect.topLeft.y, winPaint);

    }
    else

        DrawBitmap (dstWinH, bitmapID,
                    col * BitmapWidth + BoardOffsetX,
                    row * BitmapHeight + BoardOffsetY);
}


/***********************************************************************
 *
 * FUNCTION:    DrawBitmap
 *
 ***********************************************************************/

static void
DrawBitmap (WinHandle winHandle, Int16 resID, Int16 x, Int16 y)
{
    MemHandle resH;
    BitmapPtr resP;
    WinHandle currDrawWindow;

    // If passed a non-null window MemHandle,
    // set it up as the draw window, saving the current draw window

    if (winHandle)
        currDrawWindow = WinSetDrawWindow (winHandle);

    resH = DmGetResource (bitmapRsc, resID);
    ErrFatalDisplayIf (!resH, "no bitmap");
    resP = MemHandleLock (resH);
    WinDrawBitmap (resP, x, y);
    MemPtrUnlock (resP);
    DmReleaseResource (resH);

    // Restore the current draw window

    if (winHandle)
        WinSetDrawWindow (currDrawWindow);
}


/***********************************************************************
 *
 * FUNCTION:     GetPieceRectangle
 *
 ***********************************************************************/

static void
GetPieceRectangle (Int16 row, Int16 col, RectanglePtr rP)
{
    ErrFatalDisplayIf (row >= 8, "bad row");
    ErrFatalDisplayIf (col >= 8, "bad column");
    ErrFatalDisplayIf (!rP, "bad param");

    rP->topLeft.x = BoardOffsetX + BitmapWidth * col;
    rP->topLeft.y = BoardOffsetY + BitmapHeight * row;
    rP->extent.x = BitmapWidth;
    rP->extent.y = BitmapHeight;
}


/***********************************************************************
 *
 * FUNCTION:     DrawGameBoard
 *
 ***********************************************************************/

static void
DrawGameBoard (void)
{
    Int16 i, j, x1, y1, x2;
    RectangleType r;

    r.topLeft.x = 0;
    r.topLeft.y = BoardOffsetY;
    r.extent.x = BoardOffsetX - 1;
    r.extent.y = 145 - BoardOffsetY;
    WinEraseRectangle (&r, 0);

    x1 = BoardOffsetX;
    y1 = BoardOffsetY - 1;
    x2 = x1 + 8 * BitmapWidth;
    DrawBitmap (OffscreenBoardWinH, BorderTop, x1, y1);
    DrawBitmap (OffscreenBoardWinH, BorderRight, x2, y1);

    UpdateStatusDisplay ();

    for (i = 0; i < 8; i++)
        for (j = 0; j < 8; j++)
            DrawPiece (OffscreenBoardWinH, i, j);

    if (OffscreenBoardWinH) {

        RectangleType srcRect;

        srcRect.topLeft.x = BoardOffsetX;
        srcRect.topLeft.y = BoardOffsetY - 1;
        srcRect.extent.x = (8 * BitmapWidth) + 1;
        srcRect.extent.y = (8 * BitmapHeight) + 1;

        WinCopyRectangle (OffscreenBoardWinH /*srcWin */ ,
                          0 /*dstWin */ ,
                          &srcRect, BoardOffsetX, BoardOffsetY - 1, winPaint);
    }

    if (Game.first_move == true)
        DrawBitmap (0, MaskPlay, 2, 146);
    else
        DrawBitmap (0, MaskNewGame, 2, 146);

}


/***********************************************************************
 *
 * FUNCTION:     UpdateStatusDisplay
 *
 ***********************************************************************/

static void
UpdateStatusDisplay (void)
{
    Char txt1[32] = "Value(";
    Char txt2[32] = "Depth =  ";
    Char *cP;
    UInt16 txtLen;
    Int16 x, y;
    RectangleType r;
    Int32 value = 0;

    // Erase the old display, first

    r.topLeft.x = 60;
    r.topLeft.y = 0;
    r.extent.x = 100;
    r.extent.y = 13;
    WinEraseRectangle (&r, 0);

    value = Game.value;
    if (Game.draw)
        WinDrawChars ("Draw!!!", 7, 80, 0);
    else if (value > 20000) {
        WinDrawChars ("You Win!!!", 10, 80, 0);
    }
    else if (value < -20000) {
        WinDrawChars ("I Win!!!", 8, 80, 0);
    }
    else {
        cP = txt1 + 6;
        if (value < 0) {
            value = -value;
            *cP = '-';
            cP++;
        }
        StrIToA (cP, value);
        txtLen = StrLen (txt1);
        txt1[txtLen] = ')';
        txt1[++txtLen] = '\0';
        WinDrawChars (txt1, txtLen, 64, 0);
        txt2[8] = Game.depth + '0';
        txtLen = StrLen (txt2);
        WinDrawChars (txt2, txtLen, 119, 0);
    }
}


/***********************************************************************
 *
 * FUNCTION:     StartApplication
 *
 ***********************************************************************/

static UInt16
StartApplication (void)
{
    UInt16 error;
    Int16 i;

    // Initialize the random number seed;

    SysRandom (TimGetSeconds ());

    LoadGameBoard ();

    // Initialize offscreen bitmap window once (used for drawing optimization -
    // this significantly speeds up mass redraws such as during during initial
    // draw, board restoraration, and screen update)

    if (!OffscreenBitmapWinH) {
        Int16 i;

        OffscreenBitmapWinH =
            WinCreateOffscreenWindow (BitmapWidth * 15, BitmapHeight,
                                      screenFormat, &error);

        if (OffscreenBitmapWinH) {
            for (i = 0; i < 15; i++) {
                DrawBitmap (OffscreenBitmapWinH,
                            BitmapTable[i], BitmapWidth * i, 0);
            }
        }
    }

    if (!OffscreenBoardWinH) {
        OffscreenBoardWinH =
            WinCreateOffscreenWindow (BoardOffsetX + (8 * BitmapWidth) + 2,
                                      BoardOffsetY + (8 * BitmapHeight) + 2,
                                      screenFormat, &error);
    }
    return (0);
}


/***********************************************************************
 *
 * FUNCTION:    StopApplication
 *
 ***********************************************************************/

static void
StopApplication (void)
{
    STATE_PTR p;
    SUCESSOR_PTR q;

    if (r != NULL)
        free_sucessors (r);

    EmptyFreeLists ();

    while (HeadPrevStates != NULL) {
        p = HeadPrevStates;
        HeadPrevStates = HeadPrevStates->memPtr;
        MemHandleFree (p->memHandle);
    }

    while (HeadNextStates != NULL) {
        p = HeadNextStates;
        HeadNextStates = HeadNextStates->memPtr;
        MemHandleFree (p->memHandle);
    }

    SaveGameBoard ();

    FrmCloseAllForms ();

    // Free the offscreen window

    if (OffscreenBitmapWinH) {
        WinDeleteWindow (OffscreenBitmapWinH, false);
        OffscreenBitmapWinH = 0;
    }

    // Free the offscreen window

    if (OffscreenBoardWinH) {
        WinDeleteWindow (OffscreenBoardWinH, false);
        OffscreenBoardWinH = 0;
    }
}


/***********************************************************************
 *
 * FUNCTION:    MainFormDoCommand
 *
 ***********************************************************************/

static Boolean
MainFormDoCommand (UInt16 command)
{
    Boolean handled = false;
    STATE_PTR p;

    MenuEraseStatus (MenuGetActiveMenu ());

    switch (command) {

    case MainOptionsNewGameCmd:

        if (!Game.game_over && !Game.first_move) {
            if (FrmCustomAlert (NewGameAlert, NULL, NULL, NULL) == 0) {
                while (HeadNextStates != NULL) {
                    p = HeadNextStates;
                    HeadNextStates = HeadNextStates->memPtr;
                    MemHandleFree (p->memHandle);
                }
                NewGameBoard ();
                DrawGameBoard ();
            }
        }
        else {
            while (HeadNextStates != NULL) {
                p = HeadNextStates;
                HeadNextStates = HeadNextStates->memPtr;
                MemHandleFree (p->memHandle);
            }
            NewGameBoard ();
            DrawGameBoard ();
        }
        handled = true;
        break;

    case MainOptionsAboutCmd:

        FrmAlert (MainOptionAbout);
        handled = true;
        break;

    case MainOptionsSearchDepth4Cmd:

        Game.depth = 4;
        Game.temp_depth = 4;
        DrawGameBoard ();
        handled = true;
        break;

    case MainOptionsSearchDepth6Cmd:

        Game.depth = 6;
        Game.temp_depth = 6;
        DrawGameBoard ();
        handled = true;
        break;

    case MainOptionsSearchDepth8Cmd:

        Game.depth = 8;
        Game.temp_depth = 8;
        DrawGameBoard ();
        handled = true;
        break;

    }
    return handled;
}


/***********************************************************************
 *
 * FUNCTION:    MainFormInit
 *
 ***********************************************************************/

static void
MainFormInit (FormPtr frm)
{

}


/***********************************************************************
 *
 * FUNCTION:    MainFormHandleEvent
 *
 ***********************************************************************/

static Boolean
MainFormHandleEvent (EventPtr event)
{
    Int32 i;
    FormPtr frm;
    Boolean handled = false;
    STATE_PTR p, p2;
    SUCESSOR_PTR q;

    if (event->eType == ctlSelectEvent) {

        switch (event->data.ctlSelect.controlID) {

        case MainFormPlayButton:

            if (!Game.game_over && !Game.first_move) {
                if (FrmCustomAlert (NewGameAlert, NULL, NULL, NULL) == 0) {
                    while (HeadNextStates != NULL) {
                        p = HeadNextStates;
                        HeadNextStates = HeadNextStates->memPtr;
                        MemHandleFree (p->memHandle);
                    }
                    NewGameBoard ();
                    DrawGameBoard ();
                }
            }
            else if (!Game.first_move) {
                while (HeadNextStates != NULL) {
                    p = HeadNextStates;
                    HeadNextStates = HeadNextStates->memPtr;
                    MemHandleFree (p->memHandle);
                }
                NewGameBoard ();
                DrawGameBoard ();
            }
            else
                MakeComputerPlay ();

            handled = true;
            break;

        case MainFormShowMoveButton:

            for (i = 0; i < 64; ++i) {
                Game.new_square[i] = Game.old_square[i];
            }

            DrawGameBoard ();

            for (i = 0; i < 500000; i++);

            for (i = 0; i < 64; ++i) {
                Game.new_square[i] = Game.square[i];
            }

            CaptureGraphics ();

            RemoveStateFlags (&new_move);
            PostMoveRoutine ();
            if (Game.compulsory_jump_found) {
                PointerMove (Game.compulsory_jump_start_square);
                // Sets Global Variable r
            }

            DrawGameBoard ();

            Game.default_piece = NO_PIECE;

            handled = true;
            break;

        case MainFormDownButton:

            p = HeadPrevStates;
            if (p != NULL && states_equal (p, &start)) {
                handled = true;
                break;
            }
            if (p == NULL || p->memPtr == NULL || p->memPtr->memPtr == NULL) {
                if (FrmCustomAlert (NewGameAlert2, NULL, NULL, NULL) == 0) {
                    while (HeadNextStates != NULL) {
                        p = HeadNextStates;
                        HeadNextStates = HeadNextStates->memPtr;
                        MemHandleFree (p->memHandle);
                    }
                    NewGameBoard ();
                    DrawGameBoard ();
                }
                handled = true;
                break;
            }
            copy_state (p, &new_move);
            HeadPrevStates = p->memPtr;
            if (HeadNextStates == NULL) {
                p2 = HeadNextStates;
                HeadNextStates = p;
                HeadNextStates->memPtr = p2;
            }
            else if (!states_equal (HeadNextStates, &new_move)) {
                p2 = HeadNextStates;
                HeadNextStates = p;
                HeadNextStates->memPtr = p2;
            }
            else
                MemHandleFree (p->memHandle);

            p = HeadPrevStates;
            copy_state (p, &new_move);
            HeadPrevStates = p->memPtr;
            if (HeadNextStates == NULL) {
                p2 = HeadNextStates;
                HeadNextStates = p;
                HeadNextStates->memPtr = p2;
            }
            else if (!states_equal (HeadNextStates, &new_move)) {
                p2 = HeadNextStates;
                HeadNextStates = p;
                HeadNextStates->memPtr = p2;
            }
            else
                MemHandleFree (p->memHandle);

            if (states_equal (&start, &new_move)) {
                NewGameBoard ();
                DrawGameBoard ();
                handled = true;
                break;
            }
            RestoreGraphics ();
            DrawGameBoard ();

            for (i = 0; i < 500000; i++);

            p = HeadPrevStates;
            copy_state (p, &new_move);
            if (states_equal (&start, &new_move)) {
                NewGameBoard ();
                DrawGameBoard ();
                handled = true;
                break;
            }

            RemoveStateFlags (&new_move);
            PostMoveRoutine ();
            if (Game.compulsory_jump_found) {
                PointerMove (Game.compulsory_jump_start_square);
                // Sets Global Variable r
            }

            copy_state (&new_move, &old_move);
            copy_state (&new_move, &move);
            RestoreGraphics ();

            Game.value = new_move.value;
            Game.game_over = new_move.game_over;
            Game.draw = new_move.draw;

            DrawGameBoard ();

            Game.temp_depth = Game.depth;
            Game.default_piece = NO_PIECE;

            handled = true;
            break;

        case MainFormUpButton:

            p = HeadNextStates;
            if (p == NULL) {
                handled = true;
                break;
            }
            Game.first_move = false;
            copy_state (p, &new_move);
            copy_state (p, &move);
            copy_state (p, &old_move);
            HeadNextStates = p->memPtr;
            if (HeadNextStates == NULL) {
                if (r != NULL) {
                    free_sucessors (r);
                    r = NULL;
                }
                EmptyFreeLists ();
                PushState (&HeadNextStates, &new_move); // Lists contiguous
                HeadNextStates->turn = PLAYER_PIECE;
            }
            p2 = HeadPrevStates;
            HeadPrevStates = p;
            HeadPrevStates->memPtr = p2;
            RestoreGraphics ();
            DrawGameBoard ();

            for (i = 0; i < 500000; i++);

            p = HeadNextStates;
            copy_state (p, &new_move);
            HeadNextStates = p->memPtr;
            p2 = HeadPrevStates;
            HeadPrevStates = p;
            HeadPrevStates->memPtr = p2;

            RemoveStateFlags (&new_move);
            PostMoveRoutine ();
            if (Game.compulsory_jump_found) {
                PointerMove (Game.compulsory_jump_start_square);
                // Sets Global Variable r
            }

            copy_state (&new_move, &old_move);
            copy_state (&new_move, &move);
            RestoreGraphics ();

            Game.value = new_move.value;
            Game.game_over = new_move.game_over;
            Game.draw = new_move.draw;

            DrawGameBoard ();

            Game.temp_depth = Game.depth;
            Game.default_piece = NO_PIECE;

            handled = true;
            break;

        default:
            break;
        }

    }
    else if (event->eType == penDownEvent) {

        Boolean inBounds;

        handled = HandlePenDown (event->screenX, event->screenY,
                                 ((KeyCurrentState () &
                                   (keyBitPageUp | keyBitPageDown)) != 0),
                                 &inBounds);

        if (!handled && inBounds) {

            handled = true;
        }

    }
    else if (event->eType == menuEvent) {

        return MainFormDoCommand (event->data.menu.itemID);

    }
    else if (event->eType == frmUpdateEvent) {

        FrmDrawForm (FrmGetActiveForm ());
        DrawGameBoard ();
        handled = true;

    }
    else if (event->eType == frmOpenEvent) {

        frm = FrmGetActiveForm ();
        MainFormInit (frm);
        FrmDrawForm (frm);
        DrawGameBoard ();
        handled = true;

    }
    else if (event->eType == frmCloseEvent) {

    }

    return (handled);
}


/***********************************************************************
 *
 * FUNCTION:    AppHandleEvent
 *
 ***********************************************************************/

static Boolean
AppHandleEvent (EventPtr eventP)
{
    UInt16 formId;
    FormPtr frmP;

    if (eventP->eType == frmLoadEvent) {

        // Load the form resource.

        formId = eventP->data.frmLoad.formID;
        frmP = FrmInitForm (formId);
        FrmSetActiveForm (frmP);

        // Set the event handler for the form.  The handler of the currently
        // active form is called by FrmHandleEvent each time is receives an
        // event.

        switch (formId) {

        case MainForm:
            FrmSetEventHandler (frmP, MainFormHandleEvent);
            break;

        default:
            ErrNonFatalDisplay ("Invalid Form Load Event");
            break;

        }

        return true;
    }

    return false;
}


/***********************************************************************
 *
 * FUNCTION:    AppEventLoop
 *
 ***********************************************************************/

static void
AppEventLoop (void)
{
    UInt16 error;
    EventType event;

    do {

        EvtGetEvent (&event, evtWaitForever);

        // Swallow up/down arrow key chars to prevent the clicking sound
        // when the keys are being used as modifiers for marking cells

        if ((event.eType == keyDownEvent) &&
            (!TxtCharIsHardKey (event.data.keyDown.modifiers,
                                event.data.keyDown.chr)) &&
            (EvtKeydownIsVirtual (&event)) &&
            ((event.data.keyDown.chr == vchrPageUp) ||
             (event.data.keyDown.chr == vchrPageDown))) {

            ;                   // Do nothing

        }
        else if (!SysHandleEvent (&event))
            if (!MenuHandleEvent (0, &event, &error))
                if (!AppHandleEvent (&event))
                    FrmDispatchEvent (&event);

        // Check the heaps after each event

#if EMULATION_LEVEL != EMULATION_NONE
        MemHeapCheck (0);
        MemHeapCheck (1);
#endif

    } while (event.eType != appStopEvent);
}


/***********************************************************************
 *
 * FUNCTION:    PilotMain
 *
 ***********************************************************************/

UInt32
PilotMain (UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
    UInt16 error;

    if (cmd == sysAppLaunchCmdNormalLaunch) {

        error = StartApplication ();
        if (error)
            return (error);

        FrmGotoForm (MainForm);

        AppEventLoop ();
        StopApplication ();
    }

    return (0);
}


/***********************************************************************
 *
 * FUNCTION:    PointerMove
 *
 ***********************************************************************/

static void
PointerMove (Int16 i)
{
    Int16 j, k;

    RemoveStateFlags (&new_move);

    if (i < 32) {

        switch (new_move.board[i]) {

        case PLAYER_PIECE:
        case PLAYER_KING:

            new_move.piece = i;

            if (r != NULL)
                free_sucessors (r);

            r = gen_single_step_sucessors (&new_move);  // r is Global Variable

            if (r != NULL) {

                for (j = 0; j < (Int16) r->num; ++j)
                    if ((k = r->ptr[j]->piece) > 127)
                        new_move.board[k - 128] = HI_LIGHTED_MID_SQUARE;
                    else
                        new_move.board[k] = HI_LIGHTED_END_SQUARE;

                if (new_move.board[i] == PLAYER_PIECE)
                    new_move.board[i] = HI_LIGHTED_PLAYER_PIECE;
                else if (new_move.board[i] == PLAYER_KING)
                    new_move.board[i] = HI_LIGHTED_PLAYER_KING;
            }

        }

    }

}


/***********************************************************************
 *
 * FUNCTION:    RemoveStateFlags
 *
 ***********************************************************************/

static void
RemoveStateFlags (STATE_PTR p)
{
    Int16 i;

    for (i = 0; i < 32; ++i) {
        if (p->board[i] > HIGHEST_VALID_PIECE) {
            switch (p->board[i]) {

            case HI_LIGHTED_PLAYER_PIECE:

                p->board[i] = PLAYER_PIECE;
                break;

            case HI_LIGHTED_PLAYER_KING:

                p->board[i] = PLAYER_KING;
                break;

            case HI_LIGHTED_NO_PIECE:
            case HI_LIGHTED_END_SQUARE:
            case HI_LIGHTED_MID_SQUARE:
            case COMPUTER_START_SQUARE:
            case COMPUTER_JUMPED_SQUARE:

                p->board[i] = NO_PIECE;
                break;

            case COMPUTER_END_SQUARE_PIECE:

                p->board[i] = COMPUTER_PIECE;
                break;

            case COMPUTER_END_SQUARE_KING:

                p->board[i] = COMPUTER_KING;
                break;

            default:

                ErrFatalDisplay ("illegal piece in state");
            }
        }
    }
}


/***********************************************************************
 *
 * FUNCTION:    MakeComputerPlay
 *
 ***********************************************************************/

static void
MakeComputerPlay (void)
{
    Int32 i;
    STATE_PTR new, p;
    SUCESSOR_PTR q;

    if (Game.game_over)
        return;

    DrawBitmap (0, MaskMoving, 2, 146);

    if (r != NULL) {
        free_sucessors (r);
        r = NULL;
    }

    EmptyFreeLists ();

    while (HeadNextStates != NULL) {
        p = HeadNextStates;
        HeadNextStates = HeadNextStates->memPtr;
        MemHandleFree (p->memHandle);
    }

    CaptureGraphics ();
    RemoveStateFlags (&new_move);

    PushState (&HeadPrevStates, &new_move);     //List should be contiguous in memory

    if (Game.value < -20000)
        Game.temp_depth -= 2;
    else
        Game.temp_depth = Game.depth;

    Game.first_move = false;
    Game.draw = false;

    copy_state (&new_move, &move);
    copy_state (&new_move, &old_move);
    move.turn = COMPUTER_PIECE;

    q = play (&move, Game.temp_depth);

    if (q->num > 0) {
        copy_state (q->ptr[(UInt16) SysRandom (0) % q->num], &new_move);
        Game.value = -new_move.piece;
    }
    else {
        Game.game_over = true;
        if (Game.value > -20000 && Game.value < 20000) {
            Game.draw = true;
        }
    }

    free_sucessors (q);

    if (Game.value > 20000) {
        Game.temp_depth = 2;    // Won't commit suicicde

        q = play (&move, Game.temp_depth);

        if (q->num > 0) {
            copy_state (q->ptr[(UInt16) SysRandom (0) % q->num], &new_move);
        }
        else {
            Game.game_over = true;
            if (Game.value > -20000 && Game.value < 20000) {
                Game.draw = true;
            }
        }

        free_sucessors (q);
    }

    PostMoveRoutine ();

    if (!Game.compulsory_jump_found) {

        for (i=0; i<32; ++i) {

            if (old_move.board[i] != new_move.board[i]) {

                if (old_move.board[i] == COMPUTER_PIECE ||
                    old_move.board[i] == COMPUTER_KING )

                    new_move.board[i] = COMPUTER_START_SQUARE;

                else if (old_move.board[i] == PLAYER_PIECE ||
                    old_move.board[i] == PLAYER_KING )

                    new_move.board[i] = COMPUTER_JUMPED_SQUARE;

                else if (old_move.board[i] == NO_PIECE) {

                    if (new_move.board[i] == COMPUTER_PIECE)
                        new_move.board[i] = COMPUTER_END_SQUARE_PIECE;
                    else if (new_move.board[i] == COMPUTER_KING)
                        new_move.board[i] = COMPUTER_END_SQUARE_KING;

                }
            }
        }
    }

    if (r != NULL) {
        free_sucessors (r);
        r = NULL;
    }

    EmptyFreeLists ();

    new_move.value = Game.value;
    new_move.game_over = Game.game_over;
    new_move.draw = Game.draw;

    PushState (&HeadPrevStates, &new_move);     //List should be contiguous in memory

    if (Game.compulsory_jump_found)
        PointerMove (Game.compulsory_jump_start_square);        // Sets Global Variable r

    copy_state (&new_move, &move);

    RestoreGraphics ();
    UpdateStatusDisplay ();
    DrawGameBoard ();
}


/***********************************************************************
 *
 * FUNCTION:    PostMoveRoutine
 *
 ***********************************************************************/

static void
PostMoveRoutine (void)
{
    SUCESSOR_PTR q;
    Int32 i;

    q = gen_sucessors (&new_move);
    Game.player_jump_found = jump_found;
    Game.compulsory_jump_found = false;
    Game.compulsory_jump_start_square = 32;

    if (Game.player_jump_found) {       // more than one jump, could still
                                        // be compulsory if multi-jump.
        for (i = 0; i < 32; ++i) {

            if (r != NULL) {
                free_sucessors (r);
                r = NULL;
            }

            PointerMove (i);    // Sets Global Variable r

            if (r != NULL) {

                free_sucessors (r);
                r = NULL;

                if (!Game.compulsory_jump_found) {
                    Game.compulsory_jump_found = true;
                    Game.compulsory_jump_start_square = i;
                }
                else {
                    Game.compulsory_jump_found = false;
                    RemoveStateFlags (&new_move);
                    break;
                }
            }
        }
    }

    if (q->num == 0) {
        Game.game_over = true;
        if (Game.value > -20000 && Game.value < 20000) {
            Game.draw = true;
        }
    }

    free_sucessors (q);

}


/***********************************************************************
 *
 * FUNCTION:    EmptryFreeLists
 *
 ***********************************************************************/

static void
EmptyFreeLists (void)
{
    STATE_PTR p;
    SUCESSOR_PTR q;

    while (FreeSucessors != NULL) {
        q = FreeSucessors;
        FreeSucessors = FreeSucessors->memPtr;
        MemHandleFree (q->memHandle);
    }

    while (FreeStates != NULL) {
        p = FreeStates;
        FreeStates = FreeStates->memPtr;
        MemHandleFree (p->memHandle);
    }

}


/***********************************************************************
 *
 * FUNCTION:    play
 *
 ***********************************************************************/

static SUCESSOR_PTR
play (STATE_PTR p, Int16 depth)
{
    SUCESSOR_PTR q, r;
    Int16 max, t;
    Int16 i, j;

    q = gen_sucessors (p);
    r = New_SUCESSOR ();
    r->num = 0;
    max = -30000;
    for (i = 0; i < q->num; ++i) {
        t = -f (q->ptr[i], -30000, 30000, depth - 1);
        if (t == max) {
            j = r->num;
            r->ptr[j] = New_STATE ();
            copy_state_play (q->ptr[i], r->ptr[j]);
            r->ptr[j]->piece = max;
            ++(r->num);
        }
        else if (t > max) {
            max = t;
            for (j = 0; j < r->num; ++j) {
                r->ptr[j]->memPtr = FreeStates;
                FreeStates = r->ptr[j];
            }
            r->num = 1;
            r->ptr[0] = New_STATE ();
            copy_state_play (q->ptr[i], r->ptr[0]);
            r->ptr[0]->piece = max;
        }
    }
    free_sucessors (q);
    end_game_tweek (r);
    return (r);
}


/***********************************************************************
 *
 * FUNCTION:    end_game_tweek
 *
 ***********************************************************************/

static void
end_game_tweek (SUCESSOR_PTR r)
{
    Int32 value, clump_max;
    Int16 i, j;

    if (r->num < 2)
        return;

    value = r->ptr[0]->piece;

    for (i = 0; i < r->num; ++i)
        set_clump_value (r->ptr[i]);

    clump_max = r->ptr[0]->piece;

    for (i = 1; i < r->num; ++i)
        if ((j = r->ptr[i]->piece) > clump_max)
            clump_max = j;

    for (i = 0, j = 0; i < r->num; ++i) {
        if (r->ptr[i]->piece < clump_max) {
            r->ptr[i]->memPtr = FreeStates;
            FreeStates = r->ptr[i];
        }
        else {
            if (j < i)
                r->ptr[j] = r->ptr[i];
            ++j;
        }
    }

    for (i = 0; i < j; ++i)
        r->ptr[i]->piece = value;

    r->num = j;
}


/***********************************************************************
 *
 * FUNCTION:    set_clump_value
 *
 ***********************************************************************/

static void
set_clump_value (STATE_PTR p)
{
    Int32 clump;
    Int16 i, j, row1, col1, row2, col2, row_dist, col_dist;

    clump = 0;

    for (i = 0; i < 32; ++i) {

        if (p->board[i] == COMPUTER_PIECE || p->board[i] == COMPUTER_KING) {

            row1 = i / 4;
            col1 = (i % 4) * 2 + 1 - row1 + (row1 / 2) * 2;

            for (j = 0; j < 32; ++j) {

                if (p->board[j] == PLAYER_PIECE || p->board[j] == PLAYER_KING) {

                    row2 = j / 4;
                    col2 = (j % 4) * 2 + 1 - row2 + (row2 / 2) * 2;

                    if (p->board[i] == COMPUTER_PIECE)

                        clump -= (i / 4) * 16;  //First priority to king piece

                    else {      // Kings chase main clump

                        if ((row_dist = (row1 - row2)) < 0)
                            row_dist = -row_dist;
                        if ((col_dist = (col1 - col2)) < 0)
                            col_dist = -col_dist;
                        clump += 16 - row_dist - col_dist;
                    }
                }
            }
        }
    }

    p->piece = clump;
}


/***********************************************************************
 *
 * FUNCTION:    f
 *
 ***********************************************************************/

Int16
f (STATE_PTR p, Int16 alpha, Int16 beta, Int16 depth)
{
    SUCESSOR_PTR q;
    Int16 t;
    Int16 i;

    if (depth == 0)
        return (value (p));

    q = gen_sucessors (p);

    if (q->num == 0 && (alpha = value (p)) > -20000 && alpha < 20000) {
        if (!STALEMATE_LEGAL)
            alpha -= 25000;     //lose when can not move
        else if (alpha < -16)
            alpha += 16;        // seek/avoid stalemate
        else if (alpha > 16)
            alpha -= 16;        // seek/avoid stalemate
    }

    for (i = 0; i < q->num && alpha < beta; ++i) {
        t = -f (q->ptr[i], -beta, -alpha, depth - 1);
        if (t > alpha)
            alpha = t;
    }
    free_sucessors (q);
    return (alpha);
}


/***********************************************************************
 *
 * FUNCTION:    gen_sucessors
 *
 ***********************************************************************/

static SUCESSOR_PTR
gen_sucessors (STATE_PTR p)
{
    SUCESSOR_PTR q;
    STATE_PTR new;
    Int16 c, i, j, k, l;

    q = New_SUCESSOR ();
    q->num = 0;
    if (p->turn == COMPUTER_PIECE)
        t = &computer_turn_nos;
    else
        t = &oponent_turn_nos;
    jump_found = FALSE;
    for (i = 0; i < 32; ++i) {
        c = p->board[i];
        if (c == t->mover_piece || c == t->mover_king) {
            p->piece = i;
            add_jumps (q, p, FALSE);
            if (!jump_found) {
                j = num_moves[c][i];
                for (k = 0; k < j; ++k) {
                    l = moves[c][i][k] - 1;
                    if (p->board[l] == NO_PIECE) {
                        new = New_STATE ();
                        copy_state_play (p, new);
                        new->board[i] = NO_PIECE;
                        if (l < 4 || l > 27)
                            new->board[l] = t->mover_king;
                        else
                            new->board[l] = c;
                        new->turn = t->oponent_piece;
                        q->ptr[q->num] = new;
                        ++(q->num);
                    }
                }
            }
        }
    }
    return (q);
}


/***********************************************************************
 *
 * FUNCTION:    add_jumps
 *
 ***********************************************************************/

void
add_jumps (SUCESSOR_PTR q, STATE_PTR p, Int16 jumped)
{
    SUCESSOR_PTR r;
    Int16 i;

    r = consider_jumps (p);
    if (r == NULL) {
        if (jumped) {
            if (!jump_found) {
                for (i = 0; i < q->num; ++i) {
                    q->ptr[i]->memPtr = FreeStates;
                    FreeStates = q->ptr[i];
                }
                q->num = 0;
                jump_found = TRUE;
            }
            p->turn = t->oponent_piece;
            q->ptr[q->num] = p;
            ++(q->num);
        }
    }
    else {
        for (i = 0; i < r->num; ++i)
            add_jumps (q, r->ptr[i], TRUE);
        r->memPtr = FreeSucessors;
        FreeSucessors = r;
        if (jumped) {
            p->memPtr = FreeStates;
            FreeStates = p;
        }
    }
}


/***********************************************************************
 *
 * FUNCTION:    consider_jumps
 *
 ***********************************************************************/

static SUCESSOR_PTR
consider_jumps (STATE_PTR p)
{
    SUCESSOR_PTR q;
    STATE_PTR new;
    Int16 i, j, k, l, m, n, o;

    q = NULL;
    i = p->piece;
    if (i < 0)
        return (q);
    j = p->board[i];
    k = num_moves[j][i];
    for (l = 0; l < k; ++l) {
        m = moves[j][i][l];
        n = p->board[--m];
        if (n == t->oponent_piece || n == t->oponent_king) {
            o = jumps[j][i][l];
            if (o != 0 && p->board[--o] == NO_PIECE) {
                new = New_STATE ();
                copy_state_play (p, new);
                new->board[i] = NO_PIECE;
                new->board[m] = NO_PIECE;
                if (o < 4 || o > 27) {
                    new->board[o] = t->mover_king;
                    if (j == t->mover_piece)
                        new->piece = o - 64;
                    else
                        new->piece = o;
                }
                else {
                    new->board[o] = j;
                    new->piece = o;
                }
                if (q == NULL) {
                    q = New_SUCESSOR ();
                    q->num = 0;
                }
                q->ptr[q->num] = new;
                ++(q->num);
            }
        }
    }
    return (q);
}


/***********************************************************************
 *
 * FUNCTION:    free_sucessors
 *
 ***********************************************************************/

static void
free_sucessors (SUCESSOR_PTR q)
{
    Int16 i;

    for (i = 0; i < q->num; ++i) {
        q->ptr[i]->memPtr = FreeStates;
        FreeStates = q->ptr[i];
    }
    q->memPtr = FreeSucessors;
    FreeSucessors = q;
}


/***********************************************************************
 *
 * FUNCTION:    value
 *
 ***********************************************************************/

static Int16
value (STATE_PTR p)
{
    Int16 i;
    Int16 x1, x2, x3, x4;


    if (p->board[0] == PLAYER_PIECE && p->board[2] == PLAYER_PIECE)
        x1 = 2;
    else
        x1 = 0;

    if (p->board[29] == COMPUTER_PIECE && p->board[31] == COMPUTER_PIECE)
        x3 = 2;
    else
        x3 = 0;

    x2 = x4 = 0;

    for (i = 0; i < 4; ++i)
        switch (p->board[i]) {
        case PLAYER_PIECE:
            x1 += 8;
            break;
        case PLAYER_KING:
            x2 += 11;
            break;
        case COMPUTER_KING:
            x4 += 11;
        }
    for (i = 4; i < 28; ++i)
        switch (p->board[i]) {
        case PLAYER_PIECE:
            x1 += 8;
            break;
        case PLAYER_KING:
            x2 += 12;
            break;
        case COMPUTER_PIECE:
            x3 += 8;
            break;
        case COMPUTER_KING:
            x4 += 12;
        }
    for (i = 28; i < 32; ++i)
        switch (p->board[i]) {
        case PLAYER_KING:
            x2 += 11;
            break;
        case COMPUTER_PIECE:
            x3 += 8;
            break;
        case COMPUTER_KING:
            x4 += 11;
        }

    if (!x1 && !x2) {

        if (p->turn == PLAYER_PIECE)
            return( -25000 - x3 - x4 );
        else
            return(  25000 + x3 + x4 );

    } else if (!x3 && !x4) {

        if (p->turn == PLAYER_PIECE)
            return(  25000 + x1 + x2 );
        else
            return( -25000 - x1 - x2 );

    } else if (p->turn == PLAYER_PIECE)

        return ( x1 + x2 - x3 - x4 );

    else

        return ( x3 + x4 - x1 - x2 );
}


/***********************************************************************
 *
 * FUNCTION:    gen_single_step_sucessors
 *
 ***********************************************************************/

static SUCESSOR_PTR
gen_single_step_sucessors (STATE_PTR p)
{
    SUCESSOR_PTR q, r;
    STATE_PTR newp;
    Int16 c, i, j, k, l;

    q = NULL;
    p->turn = PLAYER_PIECE;
    t = &oponent_turn_nos;      // Needed for consider_jumps()

    if ((i = p->piece) > 127) {
        i -= 128;
    }
    c = p->board[i];
    if (c == PLAYER_PIECE || c == PLAYER_KING) {
        if ((q = consider_jumps (p)) != NULL) {
            for (j = 0; j < q->num; ++j) {
                if (q->ptr[j]->piece < 0) {     // Piece has Kinged through jumping
                    q->ptr[j]->piece += 64;     // Set correct square, mark as END
                    q->ptr[j]->turn = COMPUTER_PIECE;
                }
                else if ((r = consider_jumps (q->ptr[j])) != NULL) {
                    q->ptr[j]->piece += 128;    // mark as MID Square
                    free_sucessors (r);
                }
                else {
                    q->ptr[j]->turn = COMPUTER_PIECE;
                }
            }
        }
        else if (!Game.player_jump_found) {     // Global Variable
            j = num_moves[c][i];
            for (k = 0; k < j; ++k) {
                l = moves[c][i][k] - 1;
                if (p->board[l] == NO_PIECE) {
                    newp = New_STATE ();
                    copy_state (p, newp);
                    newp->board[i] = NO_PIECE;
                    if (l < 4 || l > 27)
                        newp->board[l] = PLAYER_KING;
                    else
                        newp->board[l] = c;
                    newp->piece = l;    // Mark as END Square
                    newp->turn = COMPUTER_PIECE;
                    if (q == NULL) {
                        q = New_SUCESSOR ();
                        q->num = 0;
                    }
                    q->ptr[q->num] = newp;
                    ++(q->num);
                }
            }
        }
    }
    return (q);
}


/***********************************************************************
 *
 * FUNCTION:    TrimHistory
 *
 ***********************************************************************/

static void
TrimHistory (void)
{
    STATE_PTR p, pend, pstart, pfree;
    Int8 i;

    EmptyFreeLists ();

    p = HeadPrevStates;
    for (i = 0; i < 13; ++i) {
        if (p == NULL)
            return;
        p = p->memPtr;
    }
    pend = p;
    for (i = 0; i < 15; ++i) {
        if (p == NULL)
            return;
        p = p->memPtr;
    }
    pstart = HeadPrevStates;
    while (pend != NULL) {
        pend = pend->memPtr;
        pstart = pstart->memPtr;
    }
    p = HeadPrevStates;
    HeadPrevStates = pstart;
    while (pstart != NULL) {
        copy_state (p, pstart);
        pfree = p;
        p = p->memPtr;
        pstart = pstart->memPtr;
        MemHandleFree (pfree->memHandle);
    }
    while (p != HeadPrevStates) {
        pfree = p;
        p = p->memPtr;
        MemHandleFree (pfree->memHandle);
    }
}


/***********************************************************************
 *
 * FUNCTION:    New_STATE
 *
 ***********************************************************************/

static STATE_PTR
New_STATE (void)
{
    STATE_PTR p;
    MemHandle pH;

    if (FreeStates == NULL) {
        pH = MemHandleNew (sizeof (STATE));
        if (!pH) {
            TrimHistory ();
            pH = MemHandleNew (sizeof (STATE));
            if (pH) {
                FrmAlert (MainTrimHistory);
            }
        }
        ErrFatalDisplayIf (!pH, "MemHandleNew() Returned 0, OUT OF RAM");
        p = MemHandleLock (pH);
        p->memHandle = pH;
    }
    else {
        p = FreeStates;
        FreeStates = FreeStates->memPtr;
    }
    return (p);
}


/***********************************************************************
 *
 * FUNCTION:    PushState
 *
 ***********************************************************************/

static void
PushState (STATE_PTR * r, STATE_PTR q)
{
    STATE_PTR p, s;
    MemHandle pH;
    Int8 i;

    pH = MemHandleNew (sizeof (STATE));
    if (!pH) {
        TrimHistory ();
        pH = MemHandleNew (sizeof (STATE));
        if (pH) {
            FrmAlert (MainTrimHistory);
        }
    }
    ErrFatalDisplayIf (!pH, "MemHandleNew() Returned 0, OUT OF RAM");
    p = MemHandleLock (pH);
    p->memHandle = pH;
    copy_state (q, p);
    RemoveStateFlags (p);
    p->memPtr = *r;
    *r = p;
}


/***********************************************************************
 *
 * FUNCTION:    New_SUCESSOR
 *
 ***********************************************************************/

static SUCESSOR_PTR
New_SUCESSOR (void)
{
    SUCESSOR_PTR q;
    MemHandle qH;

    if (FreeSucessors == NULL) {
        qH = MemHandleNew (sizeof (SUCESSOR));
        if (!qH) {
            TrimHistory ();
            qH = MemHandleNew (sizeof (SUCESSOR));
            if (qH) {
                FrmAlert (MainTrimHistory);
            }
        }
        ErrFatalDisplayIf (!qH, "MemHandleNew() Returned 0, OUT OF RAM");
        q = MemHandleLock (qH);
        q->memHandle = qH;
    }
    else {
        q = FreeSucessors;
        FreeSucessors = FreeSucessors->memPtr;
    }
    return (q);
}


/***********************************************************************
 *
 * FUNCTION:    copy_state
 *
 ***********************************************************************/

static void
copy_state (STATE_PTR p, STATE_PTR q)
{
    Int16 i;

    MemMove ((void *) q, (const void *) p,
             (Int32) (sizeof (STATE) -
                      sizeof (STATE_PTR) -
                      sizeof (MemHandle)));
}


/***********************************************************************
 *
 * FUNCTION:    copy_state_play
 *
 ***********************************************************************/

static void
copy_state_play (STATE_PTR p, STATE_PTR q)
{
    Int16 i;

    MemMove ((void *) q, (const void *) p,
             (Int32) (sizeof (UInt8) * 32 +
                      sizeof (Int32) +
                      sizeof (Int16)));
}


/***********************************************************************
 *
 * FUNCTION:    states_equal
 *
 ***********************************************************************/

static Boolean
states_equal (STATE_PTR p, STATE_PTR q)
{
    Int16 i;

    for (i = 0; i < 32; ++i)
        if (p->board[i] != q->board[i])
            break;
    return (i == 32);
}


/***********************************************************************
 *
 * FUNCTION:    ConvertSquare
 *
 ***********************************************************************/

static Int16
ConvertSquare (UInt16 bitmap)
{
    switch (bitmap) {

    case WhiteSquareEmptyBitmap:
        return (NO_PIECE);

    case WhiteSquareRedManBitmap:
        return (PLAYER_PIECE);

    case WhiteSquareRedKingBitmap:
        return (PLAYER_KING);

    case WhiteSquareBlackManBitmap:
        return (COMPUTER_PIECE);

    case WhiteSquareBlackKingBitmap:
        return (COMPUTER_KING);

    case HiLightedSquareEmptyBitmap:
        return (HI_LIGHTED_NO_PIECE);

    case HiLightedSquareRedManBitmap:
        return (HI_LIGHTED_PLAYER_PIECE);

    case HiLightedSquareRedKingBitmap:
        return (HI_LIGHTED_PLAYER_KING);

    case MoveEndSquareBitmap:
        return (HI_LIGHTED_END_SQUARE);

    case MoveMidSquareBitmap:
        return (HI_LIGHTED_MID_SQUARE);
    
    case ComputerStartSquareBitmap:
        return (COMPUTER_START_SQUARE);

    case ComputerEndSquareManBitmap:
        return (COMPUTER_END_SQUARE_PIECE);

    case ComputerEndSquareKingBitmap:
        return (COMPUTER_END_SQUARE_KING);

    case ComputerJumpedSquareBitmap:
        return (COMPUTER_JUMPED_SQUARE);
    }

    ErrFatalDisplayIf (1, "ilegal bitmap");
}


/***********************************************************************
 *
 * FUNCTION:    ConvertBitmap
 *
 ***********************************************************************/

static Int16
ConvertBitmap (UInt16 piece)
{
    switch (piece) {

    case NO_PIECE:
        return (WhiteSquareEmptyBitmap);

    case PLAYER_PIECE:
        return (WhiteSquareRedManBitmap);

    case PLAYER_KING:
        return (WhiteSquareRedKingBitmap);

    case COMPUTER_PIECE:
        return (WhiteSquareBlackManBitmap);

    case COMPUTER_KING:
        return (WhiteSquareBlackKingBitmap);

    case HI_LIGHTED_NO_PIECE:
        return (HiLightedSquareEmptyBitmap);

    case HI_LIGHTED_PLAYER_PIECE:
        return (HiLightedSquareRedManBitmap);

    case HI_LIGHTED_PLAYER_KING:
        return (HiLightedSquareRedKingBitmap);

    case HI_LIGHTED_END_SQUARE:
        return (MoveEndSquareBitmap);

    case HI_LIGHTED_MID_SQUARE:
        return (MoveMidSquareBitmap);
    
    case COMPUTER_START_SQUARE:
        return (ComputerStartSquareBitmap);

    case COMPUTER_END_SQUARE_PIECE:
        return (ComputerEndSquareManBitmap);

    case COMPUTER_END_SQUARE_KING:
        return (ComputerEndSquareKingBitmap);

    case COMPUTER_JUMPED_SQUARE:
        return (ComputerJumpedSquareBitmap);
    }

    ErrFatalDisplayIf (1, "ilegal piece");
}


/***********************************************************************
 *
 * FUNCTION:    CaptureGraphics
 *
 ***********************************************************************/

static void
CaptureGraphics (void)
{
    Int16 i;

    for (i = 0; i < 8; i = i + 2)
        new_move.board[3 - ((i / 2) % 4)] =
            ConvertSquare (Game.new_square[i]);
    for (i = 8; i < 16; i = i + 2)
        new_move.board[7 - ((i / 2) % 4)] =
            ConvertSquare (Game.new_square[i + 1]);
    for (i = 16; i < 24; i = i + 2)
        new_move.board[11 - ((i / 2) % 4)] =
            ConvertSquare (Game.new_square[i]);
    for (i = 24; i < 32; i = i + 2)
        new_move.board[15 - ((i / 2) % 4)] =
            ConvertSquare (Game.new_square[i + 1]);
    for (i = 32; i < 40; i = i + 2)
        new_move.board[19 - ((i / 2) % 4)] =
            ConvertSquare (Game.new_square[i]);
    for (i = 40; i < 48; i = i + 2)
        new_move.board[23 - ((i / 2) % 4)] =
            ConvertSquare (Game.new_square[i + 1]);
    for (i = 48; i < 56; i = i + 2)
        new_move.board[27 - ((i / 2) % 4)] =
            ConvertSquare (Game.new_square[i]);
    for (i = 56; i < 64; i = i + 2)
        new_move.board[31 - ((i / 2) % 4)] =
            ConvertSquare (Game.new_square[i + 1]);

    for (i = 0; i < 8; i = i + 2)
        move.board[3 - ((i / 2) % 4)] = ConvertSquare (Game.square[i]);
    for (i = 8; i < 16; i = i + 2)
        move.board[7 - ((i / 2) % 4)] = ConvertSquare (Game.square[i + 1]);
    for (i = 16; i < 24; i = i + 2)
        move.board[11 - ((i / 2) % 4)] = ConvertSquare (Game.square[i]);
    for (i = 24; i < 32; i = i + 2)
        move.board[15 - ((i / 2) % 4)] = ConvertSquare (Game.square[i + 1]);
    for (i = 32; i < 40; i = i + 2)
        move.board[19 - ((i / 2) % 4)] = ConvertSquare (Game.square[i]);
    for (i = 40; i < 48; i = i + 2)
        move.board[23 - ((i / 2) % 4)] = ConvertSquare (Game.square[i + 1]);
    for (i = 48; i < 56; i = i + 2)
        move.board[27 - ((i / 2) % 4)] = ConvertSquare (Game.square[i]);
    for (i = 56; i < 64; i = i + 2)
        move.board[31 - ((i / 2) % 4)] = ConvertSquare (Game.square[i + 1]);

    for (i = 0; i < 8; i = i + 2)
        old_move.board[3 - ((i / 2) % 4)] =
            ConvertSquare (Game.old_square[i]);
    for (i = 8; i < 16; i = i + 2)
        old_move.board[7 - ((i / 2) % 4)] =
            ConvertSquare (Game.old_square[i + 1]);
    for (i = 16; i < 24; i = i + 2)
        old_move.board[11 - ((i / 2) % 4)] =
            ConvertSquare (Game.old_square[i]);
    for (i = 24; i < 32; i = i + 2)
        old_move.board[15 - ((i / 2) % 4)] =
            ConvertSquare (Game.old_square[i + 1]);
    for (i = 32; i < 40; i = i + 2)
        old_move.board[19 - ((i / 2) % 4)] =
            ConvertSquare (Game.old_square[i]);
    for (i = 40; i < 48; i = i + 2)
        old_move.board[23 - ((i / 2) % 4)] =
            ConvertSquare (Game.old_square[i + 1]);
    for (i = 48; i < 56; i = i + 2)
        old_move.board[27 - ((i / 2) % 4)] =
            ConvertSquare (Game.old_square[i]);
    for (i = 56; i < 64; i = i + 2)
        old_move.board[31 - ((i / 2) % 4)] =
            ConvertSquare (Game.old_square[i + 1]);
}


/***********************************************************************
 *
 * FUNCTION:    RestoreGraphics
 *
 ***********************************************************************/

static void
RestoreGraphics (void)
{
    Int16 i;

    for (i = 0; i < 64; ++i)
        Game.new_square[i] = GreenSquareEmptyBitmap;
    for (i = 0; i < 8; i = i + 2)
        Game.new_square[i] =
            ConvertBitmap (new_move.board[3 - ((i / 2) % 4)]);
    for (i = 8; i < 16; i = i + 2)
        Game.new_square[i + 1] =
            ConvertBitmap (new_move.board[7 - ((i / 2) % 4)]);
    for (i = 16; i < 24; i = i + 2)
        Game.new_square[i] =
            ConvertBitmap (new_move.board[11 - ((i / 2) % 4)]);
    for (i = 24; i < 32; i = i + 2)
        Game.new_square[i + 1] =
            ConvertBitmap (new_move.board[15 - ((i / 2) % 4)]);
    for (i = 32; i < 40; i = i + 2)
        Game.new_square[i] =
            ConvertBitmap (new_move.board[19 - ((i / 2) % 4)]);
    for (i = 40; i < 48; i = i + 2)
        Game.new_square[i + 1] =
            ConvertBitmap (new_move.board[23 - ((i / 2) % 4)]);
    for (i = 48; i < 56; i = i + 2)
        Game.new_square[i] =
            ConvertBitmap (new_move.board[27 - ((i / 2) % 4)]);
    for (i = 56; i < 64; i = i + 2)
        Game.new_square[i + 1] =
            ConvertBitmap (new_move.board[31 - ((i / 2) % 4)]);

    for (i = 0; i < 64; ++i)
        Game.square[i] = GreenSquareEmptyBitmap;
    for (i = 0; i < 8; i = i + 2)
        Game.square[i] = ConvertBitmap (move.board[3 - ((i / 2) % 4)]);
    for (i = 8; i < 16; i = i + 2)
        Game.square[i + 1] = ConvertBitmap (move.board[7 - ((i / 2) % 4)]);
    for (i = 16; i < 24; i = i + 2)
        Game.square[i] = ConvertBitmap (move.board[11 - ((i / 2) % 4)]);
    for (i = 24; i < 32; i = i + 2)
        Game.square[i + 1] = ConvertBitmap (move.board[15 - ((i / 2) % 4)]);
    for (i = 32; i < 40; i = i + 2)
        Game.square[i] = ConvertBitmap (move.board[19 - ((i / 2) % 4)]);
    for (i = 40; i < 48; i = i + 2)
        Game.square[i + 1] = ConvertBitmap (move.board[23 - ((i / 2) % 4)]);
    for (i = 48; i < 56; i = i + 2)
        Game.square[i] = ConvertBitmap (move.board[27 - ((i / 2) % 4)]);
    for (i = 56; i < 64; i = i + 2)
        Game.square[i + 1] = ConvertBitmap (move.board[31 - ((i / 2) % 4)]);

    for (i = 0; i < 64; ++i)
        Game.old_square[i] = GreenSquareEmptyBitmap;
    for (i = 0; i < 8; i = i + 2)
        Game.old_square[i] =
            ConvertBitmap (old_move.board[3 - ((i / 2) % 4)]);
    for (i = 8; i < 16; i = i + 2)
        Game.old_square[i + 1] =
            ConvertBitmap (old_move.board[7 - ((i / 2) % 4)]);
    for (i = 16; i < 24; i = i + 2)
        Game.old_square[i] =
            ConvertBitmap (old_move.board[11 - ((i / 2) % 4)]);
    for (i = 24; i < 32; i = i + 2)
        Game.old_square[i + 1] =
            ConvertBitmap (old_move.board[15 - ((i / 2) % 4)]);
    for (i = 32; i < 40; i = i + 2)
        Game.old_square[i] =
            ConvertBitmap (old_move.board[19 - ((i / 2) % 4)]);
    for (i = 40; i < 48; i = i + 2)
        Game.old_square[i + 1] =
            ConvertBitmap (old_move.board[23 - ((i / 2) % 4)]);
    for (i = 48; i < 56; i = i + 2)
        Game.old_square[i] =
            ConvertBitmap (old_move.board[27 - ((i / 2) % 4)]);
    for (i = 56; i < 64; i = i + 2)
        Game.old_square[i + 1] =
            ConvertBitmap (old_move.board[31 - ((i / 2) % 4)]);
}
